#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// All tracks padded to this size
#define TRACK_SIZE 3840
void check_errors(unsigned int start, unsigned int length, 
     unsigned char *track_errors, unsigned int track)
{
   unsigned int i;
   int first_bad;
   int num_bad = 0;

   for (i = start; i < start + length; i++) {
      if (track_errors[i] != 0) {
         if (num_bad == 0)
	    first_bad = i;
         num_bad++;
      }
      if ((track_errors[i] == 0 || i == (start+length-1) ) &&
            num_bad != 0) {
         if (num_bad == 1)
            fprintf(stderr,"Track %u byte %d bad\n", track, first_bad);
         else
            fprintf(stderr,"Track %u byte %d-%d bad\n", track, first_bad,
                 first_bad + num_bad-1);
         num_bad = 0;
      }
   }
}

unsigned char bcd(unsigned char v)
{
   return (v / 10) * 16 + v % 10;
}

unsigned char *process_sector(unsigned char *sect_start_chr, unsigned char *track_data,
   unsigned char *track_errors, unsigned int remaining_length, unsigned int track, unsigned int sector) {

   unsigned char *orig_sect_start_chr;
   unsigned int sect_length, orig_length;
   int found = 0;

orig_sect_start_chr = sect_start_chr;
orig_length = remaining_length;
do {
//fprintf(stderr, "start track %d sect %d %x chr %x\n",track,sector,sect_start_chr, *sect_start_chr);
   sect_start_chr = memchr(sect_start_chr,0x76, remaining_length);
   if (sect_start_chr != NULL)
      remaining_length = orig_length - (sect_start_chr - orig_sect_start_chr);
//if (track == 60) fprintf(stderr,"found %x searching %x len %d\n",sect_start_chr, orig_sect_start_chr, remaining_length);
   if (sect_start_chr != NULL && remaining_length >= 3) {
      check_errors(sect_start_chr - track_data, 3, track_errors, track);
      sect_length = sect_start_chr[2] * 256;
//fprintf(stderr,"sector %d hdr %x length %d remaining %d\n",sector,
//         sect_start_chr[1], sect_length, remaining_length);
      if (sect_start_chr[1] == bcd(sector) &&
           sect_length <= remaining_length-3) {
         fwrite(sect_start_chr, 1, 3, stdout);
         if (sect_start_chr > orig_sect_start_chr + 10) {
            fprintf(stderr,"Warning track %d sector %d has %d extra bytes preceeding\n",
               track, sector, sect_start_chr - orig_sect_start_chr);
         }
         check_errors(sect_start_chr + 3 - track_data, sect_length, 
              track_errors, track);
         fwrite(sect_start_chr + 3, 1, sect_length, stdout);
         sect_start_chr = sect_start_chr + sect_length;
         found = 1;
      } else {
         sect_start_chr++;
      }
   }
} while (remaining_length > 0 && sect_start_chr != NULL && !found);
   if (!found) {
      sect_start_chr = NULL;
      if (orig_length > 255)
            fprintf(stderr,"Valid track %d sector %d header not found, %d bytes ignored\n",
              track, sector, orig_length);
   }
   return sect_start_chr;
}

int main()
{
   unsigned int track, sector;
   unsigned int blocks;
   unsigned int track_length, error_length;
   unsigned char *track_start_chr, *sect_start_chr;
   unsigned char track_data[16*256];
   unsigned char track_errors[16*256];
   unsigned char zeros[TRACK_SIZE];
   int i,n;
   unsigned int data, error, error_loc;
   unsigned char track_header[4] = {0x43, 0x57, 0x00, 0x58};
   unsigned int track_start;
   int size;

   memset(zeros, 0, sizeof(zeros));
   track = 0;
   while (!feof(stdin)) {
      track_start = ftell(stdout);
      n = scanf("%x",&track_length);
      if (feof(stdin)) {
         continue;	// If we hit end of file then we are done
      }
      if (n != 1) {
         fprintf(stderr, "Error reading track length track %d\n",track);
         exit(1);
      }
      if (track_length >= sizeof(track_data)) {
         fprintf(stderr, "Error, track too long %d\n",track_length);
         exit(1);
      }
      for (i = 0; i < track_length; i++) {
         n = scanf("%x",&data);
         if (n != 1) {
            fprintf(stderr, "Error reading track data track %u byte %d\n",
                track,i);
            exit(1);
         }
         track_data[i] = data; 
      }

      n = scanf("%x",&error_length);
      if (n != 1) {
         fprintf(stderr, "Error reading error length track %d\n",track);
         exit(1);
      }

      memset(track_errors, 0, sizeof(track_errors));
      for (i = 0; i < error_length; i++) {
         n = scanf("%x %x",&error, &error_loc);
         if (n != 2) {
            fprintf(stderr, "Error reading track error data track %d byte %d\n",
                track,i);
            exit(1);
         }
         if (error_loc >= sizeof(track_data)) {
            fprintf(stderr, "Error, bad error index %x\n",data);
            exit(1);
         }
         track_errors[error_loc] = 1; 
      }

 //fprintf(stderr, "Read track %d length %d error length %d writing at %ld\n",
 //    track, track_length, error_length, track_start);
      if (track == 0) {
         blocks = track_data[2];
         if (track_data[0] != 0x22 || track_data[1] != 0x00 || 
               blocks != (track_length >> 8)) {
            fprintf(stderr,"Track 0 doesn't have normal format\n");
            check_errors(0, track_length, track_errors, track);
            fwrite(track_data, 1, track_length, stdout);
         } else {
            check_errors(0, blocks*256, track_errors, track);
            fwrite(track_data, 1, 3 + blocks * 256, stdout);
         }
      } else if (track_length < 4) {
         fprintf(stderr,"Warning, track %d %d bytes long, empty track written out\n", track, track_length);
         track_header[2] = bcd(track);
         fwrite(track_header, 1, sizeof(track_header), stdout);
      } else {
         track_header[2] = bcd(track);
         fwrite(track_header, 1, sizeof(track_header), stdout);
         sector = 1;
         track_start_chr = memchr(track_data,0x43,track_length);
            /* If we didn't find the header char set to first byte */
         if (track_start_chr == NULL)
            track_start_chr = track_data;
         check_errors(track_start_chr - track_data, 4, track_errors, track);
         if (memcmp(track_start_chr, track_header, 
                    sizeof(track_header)) != 0) {
            fprintf(stderr,"Track %d header bad %d bytes ignored. Empty track written out\n",
                track, track_length);
         } else {
         if (track_start_chr - track_data > 10) {
            fprintf(stderr,"Warning, track %d start has %d extra bytes preceeding\n", 
                 track, track_start_chr - track_data);
         }
         sect_start_chr = track_start_chr + sizeof(track_header);
         while (sect_start_chr != NULL) {
            sect_start_chr = process_sector(sect_start_chr, track_data, 
                track_errors, track_length - (sect_start_chr - track_data), 
                track, sector);
            sector += 1;
         }
         }
      }
      size = ftell(stdout) - track_start;
      if (size > TRACK_SIZE)
         fprintf(stderr,"Track %d too long, should be < %d, size %d\n",track,
            TRACK_SIZE, size);
      else {
         //fprintf(stderr,"Padding with %d\n",TRACK_SIZE - size);
         fwrite(zeros, 1, TRACK_SIZE - size, stdout);
      }
      track++;
   }
   fprintf(stderr,"Converted %d tracks\n",track);

   return 0;
}
