#ifdef PC
#include "encom.h"
#else
#include <termios.h>
#include <unistd.h>
#include <memory.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define ARRAYSIZE(a) (sizeof(a) / sizeof(a[0]))
#define TRACK_SIZE 3840
#define NUM_TRACKS 77
#define FILENAME_LEN 7		// Including trailing 0
#define CPM_FILE_OFFSET 0x4997
#define FILE_OFFSET 0x3179

unsigned char   cpm_tokens[256][32] =
{
   "AND",
   "ABS",
   "ATN",
   "ASC",
   "AUTO",
   "CLOSE",
   "CONT",
   "CLEAR",
   "CINT",
   "CSNG",
   "CDBL",
   "CVI",
   "CVS",
   "CVD",
   "COS",
   "CHR$",
   "DATA",
   "DIM",
   "DEFSTR",
   "DEFINT",
   "DEFSNG",
   "DEFDBL",
   "DEF",
   "DELETE",
   "END",
   "ELSE",
   "ERASE",
   "EDIT",
   "ERROR",
   "ERL",
   "ERR",
   "EXP",
   "EOF",
   "EQV",
   "FOR",
   "FIELD",
   "FILES",
   "FN",
   "FRE",
   "FIX",
   "GOTO",
   "GO TO",
   "GOSUB",
   "GET",
   "HEX$",
   "INPUT",
   "IF",
   "INSTR",
   "INT",
   "INP",
   "IMP",
   "KILL",
   "LET",
   "LINE",
   "LOAD",
   "LSET",
   "LPRINT",
   "LLIST",
   "LPOS",
   "LIST",
   "LOG",
   "LOC",
   "LEN",
   "LEFT$",
   "LOF",
   "MERGE",
   "MOD",
   "MKI$",
   "MKS$",
   "MKD$",
   "MID$",
   "NEXT",
   "NULL",
   "NAME",
   "NEW",
   "NOT",
   "OUT",
   "ON",
   "OPEN",
   "OR",
   "OCT$",
   "PUT",
   "POKE",
   "PRINT",
   "POS",
   "PEEK",
   "READ",
   "RUN",
   "RESTORE",
   "RETURN",
   "REM",
   "RESUME",
   "RSET",
   "RIGHT$",
   "RND",
   "RENUM",
   "RESET",
   "STOP",
   "SWAP",
   "SAVE",
   "SPC(",
   "STEP",
   "SGN",
   "SQR",
   "SIN",
   "STR$",
   "STRING$",
   "SPACE$",
   "SYSTEM",
   "TRON",
   "TROFF",
   "TAB(",
   "TO",
   "THEN",
   "TAN",
   "USING",
   "USR",
   "VAL",
   "VARPTR",
   "WIDTH",
   "WAIT",
   "XOR",
   "+",
   "-",
   "*",
   "/",
   "^",
   "\\",
   "'",
   ">",
   "=",
   "<"
};

unsigned char   cpm_tokens_id[256] =
{
   247,
   6,
   14,
   21,
   171,
   195,
   154,
   146,
   28,
   29,
   30,
   43,
   44,
   45,
   12,
   22,
   132,
   134,
   173,
   174,
   175,
   176,
   152,
   170,
   129,
   162,
   166,
   167,
   168,
   214,
   215,
   11,
   47,
   250,
   130,
   192,
   198,
   211,
   15,
   31,
   137,
   137,
   141,
   193,
   26,
   133,
   139,
   218,
   5,
   16,
   251,
   200,
   136,
   177,
   196,
   201,
   158,
   159,
   27,
   147,
   10,
   48,
   18,
   1,
   49,
   197,
   252,
   50,
   51,
   52,
   3,
   131,
   150,
   199,
   148,
   213,
   157,
   149,
   191,
   248,
   25,
   194,
   153,
   145,
   17,
   23,
   135,
   138,
   140,
   142,
   143,
   169,
   202,
   2,
   8,
   172,
   204,
   144,
   165,
   203,
   212,
   209,
   4,
   7,
   9,
   19,
   216,
   24,
   189,
   163,
   164,
   208,
   206,
   207,
   13,
   217,
   210,
   20,
   220,
   161,
   151,
   249,
   242,
   243,
   244,
   245,
   246,
   253,
   219,
   239,
   240,
   241
};
int             num_cpm_tokens = 132;

typedef struct {
   unsigned char   name[FILENAME_LEN];
   int             start;
   int             end;
} DIRENT;

typedef DIRENT  DIR[64];

#define CPM_FILENAME_LEN 9	// Including trailing 0
#define CPM_FILEEXT_LEN 4	// Including trailing 0
typedef struct {
   unsigned char   name[CPM_FILENAME_LEN];
   unsigned char   ext[CPM_FILENAME_LEN];
   int             extents;
   int             len;
   int             allocations[256];
} CPM_DIRENT;

typedef CPM_DIRENT CPM_DIR[64];

unsigned char
bcd(unsigned char v)
{
   return (v / 10) * 16 + v % 10;
}

unsigned char
bcd_to_decimal(unsigned char v)
{
   return (v >> 4) * 10 + (v & 0xf);
}
dump_data(unsigned char *v)
{
   int             i;
   for (i = 0; i < 16; i++) {
      printf("%x ", *v++);
   }
   printf("\n");
}
int
read_sector(FILE * in, int track, int sector, unsigned char *buf,
	    int *more)
{
   unsigned char   track_header[4] = { 0x43, 0x57, 0x00, 0x58 };
   unsigned char   track_buf[TRACK_SIZE];
   unsigned char  *sect_start_chr;
   int             sect_length,
                   remaining_length;

   *more = 0;
   fseek(in, TRACK_SIZE * track, SEEK_SET);
   if (fread(track_buf, 1, sizeof(track_buf), in) != sizeof(track_buf)) {
      fprintf(stderr, "File read failed track %d\n", track);
      return -1;
   }

   if (track == 0) {
      if (sector == 1 && track_buf[0] == 0x22 && track_buf[1] == 0x00) {
         return track_buf[2] * 256;
      } else {
         return -1;
      }
   }

   track_header[2] = bcd(track);
   if (memcmp(track_buf, track_header, sizeof(track_header)) != 0) {
      // fprintf(stderr,"Bad track header, track %d\n", track);
      return -1;
   }
   sect_start_chr = track_buf + sizeof(track_header);
   remaining_length = TRACK_SIZE - sizeof(track_header);
   sect_start_chr = memchr(sect_start_chr, 0x76, remaining_length);
   while (sect_start_chr != NULL && remaining_length >= 3) {
      sect_length = *(sect_start_chr + 2) * 256;
      if (*(sect_start_chr + 1) == bcd(sector)) {
	 if (sect_length > TRACK_SIZE) {
	    fprintf(stderr, "Track %d sector %d length bad\n", track,
		    sector);
	    return -1;
	 }
	 memcpy(buf, sect_start_chr + 3, sect_length);
	 sect_start_chr += sect_length + 3;
	 remaining_length -= sect_length + 3;
	 sect_start_chr = memchr(sect_start_chr, 0x76, remaining_length);
	 if (sect_start_chr != NULL
	     && sect_start_chr[1] == bcd(sector + 1))
	    *more = 1;
	 return sect_length;
      }
      sect_start_chr += 3;
      sect_start_chr += sect_length;
      remaining_length -= sect_length;
   }
   // dump_data(track_buf);
   // fprintf(stderr, "Unable to find track %d sector %d\n", track,
   // sector);
   return -1;
}

void
cpm_deskew(unsigned char *buf)
{
   unsigned char   buf2[TRACK_SIZE];
   int             i;

   int             skew_tbl[] =
       { 1, 7, 13, 19, 25, 5, 11, 17, 23, 3, 9, 15, 21, 2, 8, 14, 20,
      26, 6, 12, 18, 24, 4, 10, 16, 22
   };

   for (i = 0; i < ARRAYSIZE(skew_tbl); i++) {
      memcpy(&buf2[128 * i], &buf[(skew_tbl[i] - 1) * 128], 128);
   }
   memcpy(buf, buf2, TRACK_SIZE);
}

int
get_cpm_dir(FILE * in, CPM_DIR * dir)
{
   unsigned char   buf[TRACK_SIZE];
   int             i, j, last, current, more;
   int             extent;
   int             found_match = 0;
   int             bad_entry = 0;

   if (read_sector(in, 3, 1, buf, &more) != 13 * 256) {
      //printf("CPM Directory read failed\n");
      return -1;
   }
   cpm_deskew(buf);
   last = 0;
   for (i = 0; i < ARRAYSIZE(*dir); i++) {
      if (buf[i * 32] != 0xe5) {
         if (buf[i * 32 + 15] > 0x80)
            bad_entry++;
      }
      for (j = 1; j < 12; j++) {
          if (buf[i * 32 + j] < 32 || buf[i * 32 + j] >= 128) {
             bad_entry++;
             break;
          }
      }
   }
   if (bad_entry > 3)
      return -1;

   for (i = 0; i < ARRAYSIZE(*dir); i++) {
      if (buf[i * 32] != 0xe5) {
	 memcpy((*dir)[last].name, &buf[i * 32 + 1], CPM_FILENAME_LEN - 1);
	 (*dir)[last].name[CPM_FILENAME_LEN - 1] = 0;
	 for (j = CPM_FILENAME_LEN - 2;
	      j >= 0 && (*dir)[last].name[j] == ' '; j--) {
	    (*dir)[last].name[j] = 0;
	 }
	 memcpy((*dir)[last].ext, &buf[i * 32 + 9], CPM_FILEEXT_LEN - 1);
	 (*dir)[last].ext[CPM_FILEEXT_LEN - 1] = 0;
	 for (j = CPM_FILEEXT_LEN - 2; j >= 0 && (*dir)[last].ext[j] == ' ';
	      j--) {
	    (*dir)[last].ext[j] = 0;
	 }
	 extent = buf[i * 32 + 12];
         current = last;
	 if (extent != 0) {
            found_match = 0;
            for (j = 0; j < last; j++) {
               if (strcmp((*dir)[j].name, (*dir)[last].name) == 0 &&
                    strcmp((*dir)[j].ext, (*dir)[last].ext) == 0) {
                  found_match = 1;
                  current = j;
	          if ((*dir)[current].extents + 1 != extent) {
	            printf("Extent mismatch %d %d\n", (*dir)[current].extents,
		      extent);
	          }
	          (*dir)[current].len += buf[i * 32 + 15];
                  break;
               }
            }
            if (!found_match) {
               printf("Failed to match CPM Entry %s.%s\n",(*dir)[last].name, 
                 (*dir)[last].ext); 
            }
	 } else {
	    (*dir)[last].len = buf[i * 32 + 15];
	 }
	 (*dir)[current].extents = extent;
	 for (j = 0; j < 16; j++)
	    (*dir)[current].allocations[j + extent * 16] =
		buf[i * 32 + 16 + j];
         if ((*dir)[current].extents == 0)
	    last++;
      }
   }
   return last;
}

int
get_os65d_dir(FILE * in, DIR * dir)
{
   unsigned char   buf[TRACK_SIZE];
   int             i,
                   j,
                   ndx,
                   more;

   if (read_sector(in, 8, 1, buf, &more) != 256 ||
       read_sector(in, 8, 2, &buf[256], &more) != 256) {
      //printf("OS65D Directory read failed\n");
      return -1;
   }
   ndx = 0;
   for (i = 0; i < ARRAYSIZE(*dir); i++) {
      if (buf[i * 8] != '#') {
	 memcpy((*dir)[ndx].name, &buf[i * 8], FILENAME_LEN - 1);
	 (*dir)[ndx].name[FILENAME_LEN - 1] = 0;
	 for (j = FILENAME_LEN - 2; j >= 0 && (*dir)[ndx].name[j] == ' ';
	      j--) {
	    (*dir)[ndx].name[j] = 0;
	 }
	 (*dir)[ndx].start = bcd_to_decimal(buf[i * 8 + 6]);
	 (*dir)[ndx].end = bcd_to_decimal(buf[i * 8 + 7]);
	 ndx++;
      }
   }
   return ndx;
}

unsigned char   default_tokens[][16] = {
   // "*ERR*",
   "END",
   "FOR",
   "NEXT",
   "DATA",
   "INPUT",
   "DIM",
   "READ",
   "LET",
   "GOTO",
   "RUN",
   "IF",
   "RESTORE",
   "GOSUB",
   "RETURN",
   "REM",
   "STOP",
   "ON",
   "NULL",
   "WAIT",
   "EXIT",
   "DISK",
   "DEF",
   "POKE",
   "PRINT",
   "CONT",
   "LIST",
   "CLEAR",
   "NEW",
   "TAB(",
   "TO",
   "FN",
   "SPC(",
   "THEN",
   "NOT",
   "STEP",
   "+",
   "-",
   "*",
   "/",
   "^",
   "AND",
   "OR",
   ">",
   "=",
   "<",
   "SGN",
   "INT",
   "ABS",
   "USR",
   "FRE",
   "POS",
   "SQR",
   "RND",
   "LOG",
   "EXP",
   "COS",
   "SIN",
   "TAN",
   "ATN",
   "PEEK",
   "LEN",
   "STR$",
   "VAL",
   "ASC",
   "CHR$",
   "LEFT$",
   "RIGHT$",
   "MID$"
};
// }

decode_basic_file(FILE * out, unsigned char *buf, int len)
{
   int             i;
   unsigned char   c;
   int             ptr;
   int             start_file,
                   end_file,
                   last_ptr;
   int             last_line = -1;
   int             mismatch = 0;
   int             bad_basic_tokens = 0;

   start_file = buf[1] * 256 + buf[0];
   end_file = buf[3] * 256 + buf[2];
   // printf("First %x last %x\n",start_file, end_file);
   last_ptr = start_file;
   i = start_file - FILE_OFFSET;
   while (i < len) {
      ptr = buf[i + 1] * 256 + buf[i];
      if (ptr == 0)
	 break;
      if (i + FILE_OFFSET != last_ptr)
	 if (mismatch++ < 5)
	    printf("   Basic pointer mismatch after line %d %x %x\n",
		   last_line, last_ptr, i + FILE_OFFSET);
	 else if (mismatch++ == 6)
	    printf("   Rest of mismatches ignored\n");
      last_ptr = ptr;

      // fprintf(out,"%x %x ",ptr);
      i = i + 2;
      last_line = buf[i + 1] * 256 + buf[i];
      fprintf(out, "%d ", last_line);
      i = i + 2;
      for (; i < len && buf[i] != 0; i++) {
	 if (buf[i] & 0x80) {
	    c = buf[i] & 0x7f;
	    if (c > ARRAYSIZE(default_tokens))
	       //printf("   Bad basic token %x at %d\n", buf[i], i);
               bad_basic_tokens++;
	    else
	       fprintf(out, "%s", default_tokens[c]);
	 } else
	    fprintf(out, "%c", buf[i]);
      }
      fprintf(out, "\n");
      i++;
   }
   i += 3;
   if (i + FILE_OFFSET != end_file)
      printf("   Basic file end mismatch %x %x\n", end_file,
	     i + FILE_OFFSET);
   if (bad_basic_tokens != 0)
      printf("   %d bad basic tokens\n", bad_basic_tokens);
}

decode_cpm_basic_tokens(unsigned char *buf, int len)
{
   int             i;
   char            first_letter = 'A';
   int             new_token = 2;
   unsigned char   token[sizeof(cpm_tokens[0])];
   int             ndx;
   unsigned char   token_id;


   ndx = 0;
   num_cpm_tokens = 0;
   for (i = 315; i < 1000; i++) {
      if (ndx >= sizeof(token) - 1) {
	 printf("Basic token overflow\n");
	 ndx = 0;
      }
      if (buf[i] == 0) {
	 first_letter++;
	 if (first_letter == '\\')
	    break;
      } else if (new_token == 1) {
	 cpm_tokens_id[num_cpm_tokens] = buf[i];
	 num_cpm_tokens++;
	 new_token = 2;
      } else if (buf[i] >= 0xA0) {
	 if (new_token) {
	    ndx = 0;
	    if (first_letter <= 'Z') {
	       token[ndx++] = first_letter;
	    }
	    new_token = 0;
	 }
	 new_token = 1;
	 token[ndx++] = buf[i] & 0x7f;
	 token[ndx] = 0;
	 strcpy(cpm_tokens[num_cpm_tokens], token);
      } else {
	 if (new_token) {
	    ndx = 0;
	    if (first_letter <= 'Z') {
	       token[ndx++] = first_letter;
	    }
	    new_token = 0;
	 }
	 token[ndx++] = buf[i] & 0x7f;
      }
   }
}

decode_cpm_basic_file(FILE * out, unsigned char *buf, int len)
{
   int             i;
   unsigned char   c;
   int             ptr;
   int             last_ptr;
   int             last_line = -1;
   int             mismatch = 0;
   unsigned char  *loc;
   unsigned char   chr;
   double          dv;

   last_ptr = -1;
   i = 1;
   while (i < len) {
      ptr = buf[i + 1] * 256 + buf[i];
      if (i + CPM_FILE_OFFSET != last_ptr && last_ptr != -1)
	 if (mismatch++ < 5)
	    printf("   Basic pointer mismatch after line %d %x %x\n",
		   last_line, last_ptr, i + CPM_FILE_OFFSET);
	 else if (mismatch++ == 6)
	    printf("   Rest of mismatches ignored\n");
      last_ptr = ptr;

      i = i + 2;
      last_line = buf[i + 1] * 256 + buf[i];
      fprintf(out, "%d ", last_line);
      i = i + 2;
      for (; i < len && buf[i] != 0; i++) {
	 if (buf[i] == 29) {
#if 0
	    // fprintf(out,"%x",buf[i+1] + (buf[i+2] << 8) );
	    {
	       int             z;
	       for (z = i; z < i + 10; z++) {
		  fprintf(out, "%x ", buf[z]);
	       }
	    }
#endif
	    dv = (buf[i + 3] | 0x80) / 256.0 + buf[i + 2] / (256.0 * 256) +
		buf[i + 1] / (256.0 * 256 * 256);
	    if (buf[i + 3] & 0x80)
	       dv = -dv;
	    dv = dv * pow(2, buf[i + 4] - 0x80);
	    fprintf(out, "%g", dv);

	    i += 4;
	 } else if (buf[i] == 14) {
	    fprintf(out, "%d", buf[i + 1] + (buf[i + 2] << 8));
	    i += 2;
	 } else if (buf[i] == 28) {
	    fprintf(out, "%d", buf[i + 1] + (buf[i + 2] << 8));
	    i += 2;
	 } else if (buf[i] == 15) {
	    fprintf(out, "%d", buf[i + 1]);
	    i++;
	 } else if (buf[i] >= 17 && buf[i] <= 26) {
	    fprintf(out, "%d", buf[i] - 17);
	 } else if (buf[i] & 0x80) {
	    chr = buf[i];
	    if (buf[i] == 0xff) {
	       i++;
	       chr = buf[i] & 0x7f;
	    }
	    loc = memchr(cpm_tokens_id, chr, num_cpm_tokens);
	    if (loc == NULL) {
	       printf("   Bad basic token %x at %d\n", buf[i], i);
	    } else
	       fprintf(out, "%s", cpm_tokens[loc - cpm_tokens_id]);
	 } else
	    fprintf(out, "%c", buf[i]);
      }
      fprintf(out, "\n");
      i++;
   }
   i += 3;
}

decode_asm_file(FILE * out, unsigned char *buf, int len)
{
   int             i,
                   j;
   unsigned char   c;
   int             line;
   int             start_file,
                   end_file;
   int             last_line = -1;

   start_file = buf[1] * 256 + buf[0];
   end_file = buf[3] * 256 + buf[2];
   // printf("First %x last %x\n",start_file, end_file);
   i = start_file - FILE_OFFSET;
   while (i < len && i < (end_file - FILE_OFFSET)) {
      line = buf[i + 1] * 256 + buf[i];
      if (line != 0)
	 fprintf(out, "%-4d", line);
      i = i + 2;
      for (; i < len && buf[i] != '\r'; i++) {
	 if (line != 0) {
	    if (buf[i] < 0x80)
	       fprintf(out, "%c", buf[i]);
	    else
	       for (j = 0; j < 0x100 - buf[i]; j++)
		  fprintf(out, "%c", buf[i - 1]);
	 }
      }
      if (line != 0)
	 fprintf(out, "\n");
      i++;
   }
   if (i + FILE_OFFSET != end_file)
      printf("   ASM file end mismatch %x %x\n", end_file,
	     i + FILE_OFFSET);
}
int
is_basic_file(unsigned char buf[], unsigned int len, unsigned int first)
{
   unsigned int    line1,
                   line2;
   unsigned int    second;
   int             i;
   int             token_found = 0;
   int             bad_token = 0;

   if (first >= len || first < 4)
      return 0;
   second = buf[first + 1] * 256 + buf[first] - FILE_OFFSET;
   line1 = buf[first + 3] * 256 + buf[first + 2];
   // printf("Second %d len %d", second, len);
   if (second >= len || second < 6) {
      // Sometimes the head pointer is messed up. See if we find
      // something that looks like basic
      for (i = first + 4; buf[i] != 0 && i < len - 5; i++) {
	 if (buf[i] >= 128) {
	    token_found = 1;
	    if ((buf[i] - 128) > ARRAYSIZE(default_tokens))
	       bad_token = 1;
	 }
      }
      // printf("Bad %d found %d buf %d\n",bad_token, token_found,
      // buf[i]);
      if (bad_token || !token_found || buf[i] != 0)
	 return 0;
      else
	 second = i + 1;
   }

   line2 = buf[second + 3] * 256 + buf[second + 2];

   // printf("line 1 %d line 2 %d buf %d\n",line1, line2, buf[second-1]);
   if (line2 > line1 && buf[second - 1] == 0)
      return 1;
   else
      return 0;
}

int
is_asm_file(unsigned char buf[], unsigned int len, unsigned int first)
{
   unsigned int    line1,
                   line2;
   unsigned char  *second;

   if (first + 2 >= len || len < 6)
      return 0;

   line1 = buf[first + 1] * 256 + buf[first];
   second = memchr(&buf[first + 2], 0xd, len - first - 2);
   if (second == NULL || (second - buf + 2) >= len)
      return 0;

   line2 = second[2] * 256 + second[1];
   if (first < 16 && (second - buf) < 150 && line2 > line1)
      return 1;
   else
      return 0;
}

get_file(FILE * in, DIRENT dirent)
{
   unsigned char   buf[TRACK_SIZE * NUM_TRACKS];
   int             trk,
                   sector,
                   more,
                   size,
                   ndx;
   FILE           *out;
   unsigned char   name[FILENAME_LEN];
   unsigned char  *loc;
   int             first;
   int             printed = 0;

   strcpy(name, dirent.name);
   while (loc = strchr(name, '/')) {
      *loc = '\\';
   }
   out = fopen(name, "w");
   if (out == NULL) {
      fprintf(stderr, "Open failed on %s ", name);
      perror("");
      exit(1);
   }
   ndx = 0;
   for (trk = dirent.start; trk <= dirent.end; trk++) {
      more = 1;
      sector = 1;
      do {
	 size = read_sector(in, trk, sector, &buf[ndx], &more);
	 // printf("read %d trk %d sector %d more %d\n",size, trk, sector, 
	 // more);
	 if (size < 0) {
	    if (sector == 1) {
	       if (printed++ == 0)
		  printf("\n");
	       printf("   Error reading file %s track %d sector %d\n",
		      dirent.name, trk, sector);
	    }
	 } else {
	    // fwrite(&buf[ndx], 1, size, out);
	    ndx = ndx + size;
	 }
	 sector++;
      } while (more);
   }
   first = buf[1] * 256 + buf[0] - FILE_OFFSET;
   // printf(" %d seconda %x %d\n",first, seconda, seconda-buf);

   if (is_basic_file(buf, ndx, first)) {
      printf("   BASIC\n");
      decode_basic_file(out, buf, ndx);
   } else if (is_asm_file(buf, ndx, first)) {
      printf("   ASM\n");
      decode_asm_file(out, buf, ndx);
      // fwrite(buf, 1, ndx, out);
   } else {
      printf("   Other\n");
      fwrite(buf, 1, ndx, out);
   }

   fclose(out);
}
int
is_cpm_text_file(unsigned char buf[], unsigned int len)
{
   int i;
   for (i = 0; i < len; i++) {
      if (buf[i] >= 0x80)
         return 0;
   }
   return 1;
}

int
get_cpm_file(FILE * in, CPM_DIRENT dirent, unsigned char *file_buf)
{
   unsigned char   buf[TRACK_SIZE * 2];
   int             trk,
                   more,
                   size,
                   ndx;
   int             printed = 0;
   int             i,
                   offset;

   ndx = 0;
   for (i = 0; i < dirent.len / 8; i++) {
      trk = dirent.allocations[i] * 1024 / 3328 + 3;
      offset = (dirent.allocations[i] * 1024) % 3328;
      size = read_sector(in, trk, 1, &buf[0], &more);
      cpm_deskew(buf);
      if (size != 3328) {
	 if (printed++ == 0)
	    printf("\n");
	 printf("   Error reading file %s track %d\n", dirent.name, trk);
	 continue;
      }
      if (offset + 1024 > 3328) {
	 size = read_sector(in, trk + 1, 1, &buf[3328], &more);
	 cpm_deskew(&buf[3328]);
      }
      if (size != 3328) {
	 if (printed++ == 0)
	    printf("\n");
	 printf("   Error reading file %s track %d\n", dirent.name, trk);
	 continue;
      }
      {
	 memcpy(&file_buf[ndx], &buf[offset], 1024);
	 ndx = ndx + 1024;
      }
   }
#if 0
   first = buf[1] * 256 + buf[0] - FILE_OFFSET;
   // printf(" %d seconda %x %d\n",first, seconda, seconda-buf);

   if (is_basic_file(buf, ndx, first)) {
      printf("   BASIC\n");
      decode_basic_file(out, buf, ndx);
   } else if (is_asm_file(buf, ndx, first)) {
      printf("   ASM\n");
      decode_asm_file(out, buf, ndx);
      // fwrite(buf, 1, ndx, out);
   } else {
      printf("   Other\n");
      fwrite(buf, 1, ndx, out);
   }
#endif

   return dirent.len * 128;
}

main(int argc, char *argv[])
{
   int             fd;
   FILE           *in;
   char            filename[256];
   unsigned char   file_buf[TRACK_SIZE * NUM_TRACKS];
   int             file_len;
   unsigned char   buf[TRACK_SIZE];
   int             i,j;
   DIR             dir;
   CPM_DIR         cpm_dir;
   int             dir_size;
   FILE           *out;
   unsigned char   name[CPM_FILENAME_LEN + CPM_FILEEXT_LEN];
   unsigned char  *loc;

   if (argc > 1) {
      strcpy(filename, argv[1]);
   } else {
      printf("Enter image to extract files from\n");
      fflush(stdout);
      scanf("%s", filename);
   }

#ifdef PC
   in = fopen(filename, "rb");
#else
   in = fopen(filename, "r");
#endif
   if (in == NULL) {
      fprintf(stderr, "On file %s ", filename);
      perror("open failed");
      exit(1);
   }

   dir_size = get_os65d_dir(in, &dir);
   if (dir_size == -1) {
      dir_size = get_cpm_dir(in, &cpm_dir);
      if (dir_size == -1) {
         int more, size;
        
         printf("Assuming UCSD P system directory\n");
         fflush(stdout);
         out = fopen("raw_image", "wb");
         for (i = 1; i < 77; i++) {
            if (i == 12 || i == 40 || i == 64)
               continue;

            size = read_sector(in, i, 1, buf, &more);
            if (size == -1) {
               memset(buf, 0, sizeof(buf));
               fwrite(buf, 1, 13*256, out);
            } else {
               int z, found;
	       //cpm_deskew(buf);
               fwrite(buf, 1, size, out);
#if 0
               for (j = 0; j < size; j += 256) {
                  found = 0;
                  for (z = j; z < j + 256; z++) {
                     if ((buf[z] < '0' || buf[z] > '9') && buf[z] != '\r'
                          && buf[z] != ' ' && buf[z] != 0) {
                        printf("Char %x\n",buf[z]);
                        found = 1;
                        break;
                     }
                  } 
                  if (found)
                     printf("Track %d sector %d not data\n",i, j / 256);
               }
#endif
            }
            //printf("Track %d size %d more %d\n", i, size, more);
         }
         fclose(out);
         system("/home/djg/portb/ucsd/td raw_image");
         system("/home/djg/portb/ucsd/bu raw_image > convert.txt");
      } else {
         printf("CPM Directory\n");
      }
      for (i = 0; i < dir_size; i++) {
	 if (strcmp(cpm_dir[i].name, "BASIC") == 0 &&
	     strcmp(cpm_dir[i].ext, "COM") == 0) {
	    file_len = get_cpm_file(in, cpm_dir[i], file_buf);
	    decode_cpm_basic_tokens(file_buf, file_len);
#if 0
            for (j = 0; j < num_cpm_tokens; j++) {
               printf("   \"%s\",\n",cpm_tokens[j]);
            }
            for (j = 0; j < num_cpm_tokens; j++) {
               printf("   %d,\n", cpm_tokens_id[j]);
            }
#endif
	    break;
	 }
      }
      for (i = 0; i < dir_size; i++) {
	 printf("%-8s.%-3s %3d  ", cpm_dir[i].name, cpm_dir[i].ext,
		cpm_dir[i].len);
	 file_len = get_cpm_file(in, cpm_dir[i], file_buf);
	 strcpy(name, cpm_dir[i].name);
	 strcat(name, ".");
	 strcat(name, cpm_dir[i].ext);
	 while (loc = strchr(name, '/')) {
	     *loc = '\\';
	 }
	 out = fopen(name, "w");
	 if (out == NULL) {
	    fprintf(stderr, "Open failed on %s ", name);
	    perror("");
	    exit(1);
	 }
	 if (strcmp(cpm_dir[i].ext, "BAS") == 0) {
            printf("BASIC\n");
	    decode_cpm_basic_file(out, file_buf, file_len);
	 } else if (is_cpm_text_file(file_buf, file_len)) {
            printf("Text\n");
            for (j = 0; j < file_len; j++) {
               if (file_buf[j] == 13 && file_buf[j+1] == 10) {
                  fwrite(&file_buf[j+1], 1, 1, out);
                  j++;
               } else
                  fwrite(&file_buf[j], 1, 1, out);
               if (file_buf[j] == 26)
                  break;
            }
         } else {
            printf("Other\n");
	    fwrite(file_buf, 1, file_len, out);
         }
	 fclose(out);
      }
   } else {
      printf("OS65D Directory\n");
      for (i = 0; i < dir_size; i++) {
	 printf("%-6s %2d %2d", dir[i].name, dir[i].start, dir[i].end);
	 if (dir[i].start > 0)	// Don't try to extract system
	    get_file(in, dir[i]);
	 else
	    printf("\n");
      }
   }
}
