;
; RAW DISK DUMP UTILITY FOR C1P/UK101
;
; DUMPS THE CONTENTS OF A DISK AND SENDS OUTPUT
; TO THE CASSETTE PORT AS ASCII HEX. IT WILL
; CAPTURE ALL DATA ON THE DISK INCLUDING THE
; TRACK/SECTOR FORMATTING INFORMATION.
;
; DUMPS EVERY BYTE ENCOUNTERED ON A TRACK
; STARTING AT THE INDEX PULSE.
;
; NOTE: DOS FORMATTED TRACKS WILL CONTAIN
; SPURIOUS CHARACTERS BEFORE AND AFTER SECTORS
; AND CAN BE IGNORED.
;
; MAXTRK MUST BE SET PROPERLY FOR 5 1/4" OR 8"
;
; HISTORY:
;
; 0.5  MODIFIED TO DUMP 8" DISKS AND REPORT ERRORS
;      DAVID GESSWEIN
;
; 0.4  CODE REVAMPED. DISABLED ERROR CHECK
;      (USELESS FOR DOS DISKS). REMOVED
;      MESSAGES FROM DUMP OUTPUT.
;
; 0.3  INCREASED DELAYS FOR USE ON FASTER
;      MACHINES.
;
; 0.21 FIXED ERROR COUNT. CHANGED BUFFER
;      BUFFER SIZE TO 9 PAGES AND FILL CHAR
;      TO $00.
;
; 0.2  ACIA IS RESET BEFORE EACH TRACK READ.
;      PARITY ERRORS CHECKED. READING STOPS
;      AT THE NEXT INDEX PULSE, SO NO MORE
;      EXTRANEOUS DATA IN TRACK DUMPS.
;
; 0.1  1ST RELEASE
;

ORG	=	$0400

MAXPAG	=	32	; # PAGES TO READ, 16 DATA, 16 STATUS
FILLCH	=	$00	; FILL CHARACTER

BUFFER	=	$4000	; TRACK & STATUS BUFFER

; OSI MONITOR ROM ROUTINES

LOADF	=	$203	; LOAD FLAG
SAVEF	=	$205	; SAVE FLAG
INPUT	=	$FEED	; GET CHAR FROM KEYBOARD/ACIA
;OUTPUT	=	$FFEE	; PUT CHAR TO SCREEN/ACIA

DISK	=	$C000	; DISK CONTROLLER
ACIACTL	=	$FC00	; SERIAL CONTROL REGISTER

CR	=	$0D
LF	=	$0A

NDX 	= 	$FD	; INDIRECT INDEX MEMORY LOCATION
CNTR	=	$FF	; GENERAL COUNTER

	*=	ORG
	CLD
	JSR	SAVEOFF
	JSR	PRINT
	.BYTE	CR,LF
	.BYTE	'RAW DISK DUMP 0.5',CR,LF,CR,LF
	.BYTE	'DUMP RAW TRACK DATA FROM',CR,LF
	.BYTE	'DISKS TO CASSETTE',CR,LF,CR,LF,0
	JSR	PRINT
	.BYTE	'WRITE PROTECT ALL DISKS!',CR,LF,0
	LDA	#0
	STA	BINARY
	LDA	#77
	STA	MAXTRK
	LDA	#$17
	STA	ACIACTL
	LDA	#$15
	STA	ACIACTL
CONF1	JSR	PRINT
	.BYTE	CR,LF,CR,LF
        .BYTE 	'Enter your disk drive type',CR,LF
	.BYTE 	'(8) inch or (5).25 inch? >',0
	JSR	INPUT
	JSR 	OUTPUT
	CMP	#'8'
	BEQ	SETERR
	CMP	#'5'
	BNE	CONF1
	LDA	#40
	STA	MAXTRK
SETERR	JSR 	PRINT
	.BYTE	CR,LF,CR,LF
	.BYTE 	'Select drive',CR,LF
	.BYTE 	'(A), (B), (C), (D) ? >',0
	JSR 	INPUT
	JSR 	OUTPUT
	CMP 	#$60
	BCC 	*+4
	EOR 	#$20
	CMP 	#'A
	BCC 	SETERR
	CMP 	#'E
	BCS 	SETERR
	SBC 	#'@		;carry is clear so val -1
	STA 	DRVNUM
CONF2	JSR	PRINT
	.BYTE	CR,LF,CR,LF
	.BYTE	'Select to dump in',CR,LF
	.BYTE 	'(A)scii or (B)inary ? >',0
	JSR	INPUT
	JSR 	OUTPUT
	ORA	#$20
	CMP	#'a'
	BEQ	CONF3
	CMP	#'b'
	BNE	CONF2
	INC	BINARY
CONF3	JSR	PRINT
	.BYTE	CR,LF,CR,LF
        .BYTE 	'Set serial port',CR,LF
	.BYTE	'(N)ormal or 16x (F)aster? >',0
	JSR	INPUT
	JSR 	OUTPUT
	ORA	#$20
	CMP	#'n'
	BEQ	START
	CMP	#'f'
	BNE	CONF3
	LDA	#$17
	STA	ACIACTL
	LDA	#$14
	STA	ACIACTL
START	JSR	SAVEOFF
	LDA	#0
	STA	TRK
	STA	READCNT
	LDA	MAXTRK
	STA	ENDTRK
	JSR	INITPIA
	JSR	SELDRV
	JSR	PRINT
	.BYTE	CR,LF,CR,LF
	.BYTE	'Insert disk to be dumped. Hit T',CR,LF
	.BYTE	'to select track or return for',CR,LF
	.BYTE	'entire disk',0
	JSR	INPUT
	JSR	OUTPUT
	ORA	#$20
	CMP	#'t'	
	BNE	GOZERO
	JSR	GETTRK
	INC	ENDTRK
	LDA	TRK
	CLC
	ADC	#1
	CMP	ENDTRK
	BNE	GOZERO
	LDA	#8	; IF ONLY ONE TRACK READ 8 TIMES
	STA	READCNT
GOZERO	JSR	TZERO	; STEP TO TRACK 0
	LDA	TRK
	STA	CNTR
SEEK    DEC     CNTR
        BMI     LOOP3
        JSR     TNEXT   ; STEP TO STARTING TRACK
	JMP	SEEK


LOOP	LDA	TRK
	CMP	ENDTRK
	BNE	CHKCNT
	JMP	QUIT
CHKCNT	LDA	READCNT
	BEQ	NTRK
TEST	CMP	#4
	BNE	DECRCNT
	JSR	TNEXT
	JSR	TPREV	
DECRCNT	DEC	READCNT
	JMP	LOOP3
NTRK	JSR	TNEXT	; STEP TO NEXT TRACK
LOOP3	LDA	TRK
	JSR	READMSG
	JSR	RDTRK
	LDA	#$20
	JSR	OUTPUTC
	LDA	NDX+1
	SEC
	SBC	#>BUFFER
	LSR	A	; DIVIDE BY 2 TO GET BYTE READ
	PHP
	STA	CNT+1
	JSR 	PHEXA	; PRINT BYTES READ HIGH
	TYA
	LSR	A
	PLP		; ROTATE LSB FROM HIGH INTO LOW
	BCC	LOOP4
	ORA	#$80
LOOP4	STA	CNT
	JSR	PHEXA	; AND LOW
	JSR	SAVMSG	; PRINT SAVING MESSAGE
	JSR	SAVBUF	: AND SEND DATA OUT SERIAL PORT
	LDA	READCNT
	BNE	JLOOP
	INC	TRK
JLOOP	JMP	LOOP

;SELECT DRIVE PB5 PA6  DRIVE (1-4)
;              0   0	#4
;              0   1	#3
;              1   0	#2
;              1   1	#1

SELDRV 	LDA DRVNUM
	LSR A
	TAY
	BCC *+5
	LDA #$00
	BIT $40A9
	STA DISK
	LDA DISK+2
	ORA #$20
	CPY #$01
	BCC	*+4
	AND #$DF
	STA DISK+2
	RTS
QUIT
	JSR	PRINT
	.BYTE	CR,LF,CR,LF
	.BYTE	'DONE ... REMOVE DISK',CR,LF,0
	JSR	INPUT
	JMP	START

; SAVE BUFFER CONTENTS (A)

SAVBUF	JSR	SAVEON	; SEND TO SERIAL PORT
	JSR	RESERR	; CLEAR ERROR COUNTER
	LDA	#0
	STA	ERRDMP
	LDA	CNT+1	; PRINT NUMBER OF DATA BYTES READ
	JSR 	PHEXA
	LDA	CNT
	JSR	PHEXA
	JSR	RESPTR
	LDA	CNT
	PHA		;SAVE CNT
	LDA	CNT+1
	PHA
	ORA	CNT
	BNE	SAVBUF0	;IF CNT 0 WE ARE DONE
	JMP	SAVBUF5
SAVBUF0	LDA	#0
	TAY		;BUFFER INDEX
	TAX		;LINE LENGTH COUNTER
	LDA	#1	;FORCE INITIAL CRLF ETC
	STA	SENT
SAVBUF1	LDA	CNT+1
	ORA	CNT
	BEQ	SAVBUF5	;IF CNT 0 WE ARE DONE
	LDA	SENT	;IF SENT=0 DON'T NEED SPACE ETC
	BEQ	SAVBUF3
	LDA	#0
	STA	SENT
	CPX	#0
	BNE	SAVBUF2
	JSR	CRLF
	LDX	#16
	BNE	SAVBUF3
SAVBUF2	LDA	#' '
	JSR	OUTPUTC
SAVBUF3	LDA	(NDX),Y
	AND	#$70	;REPORT PARITY OR OVERRUN ERRORS
	BEQ	NOERR
	BIT	ERRDMP
	BPL	INCERR
	JSR	PHEXA	;PRINT ERROR VALUE
	LDA	#' '
	JSR 	OUTPUTC
	LDA	NDX+1	;DUMPING ERRORS, SEND OFFSET
	SEC
	SBC	#>BUFFER
	LSR	A
	PHP
	JSR	PHEXA
	DEX
	TYA
	LSR	A
	PLP
	BCC	SAVCC
	ORA	#$80
SAVCC	JSR	PHEXA
	DEX
	INC	SENT	;MARK WE SENT ERROR
INCERR	INC	ERRCNT
	BNE	NOERR
	INC	ERRCNT+1
NOERR	INY
	BIT	ERRDMP	;DUMPING ERRORS, SKIP DATA
	BMI	DECCNT
	LDA	(NDX),Y
	JSR	PHEXA
	DEX
	INC	SENT
DECCNT	DEC	CNT
	LDA	#$FF
	CMP	CNT
	BNE	SAVBUF4
	DEC	CNT+1
SAVBUF4	INY
	BNE	SAVBUF1
	INC	NDX+1
	JMP	SAVBUF1
SAVBUF5 JSR	CRLF
	BIT	ERRDMP	;IF WE DUMPED ERRORS EXIT
	BPL	SETDMP
SAVBUF6	JSR	CRLF
	JMP	SAVEOFF	;EXIT ROUTINE
SETDMP	DEC	ERRDMP	;FLAG DUMPING ERRORS, SET TO FF
	JSR	RESPTR
	PLA
	STA	CNT+1
	PLA
	STA	CNT
	JSR 	SAVEOFF
	LDA	#' '
	JSR	OUTPUT
	LDA	ERRCNT+1;SEND NUMBER OF ERRORS TO SCREEN
	JSR 	PHEXA	; PRINT BYTES READ HIGH
	LDA	ERRCNT
	JSR	PHEXA
	JSR 	SAVEON
	LDA	ERRCNT+1;SEND NUMBER OF ERRORS TO SERIAL
	JSR	PHEXA
	LDA	ERRCNT
	JSR	PHEXA
	LDA	ERRCNT
	ORA	ERRCNT+1
	BEQ	SAVBUF7	;IF NO ERRORS EXIT
	JMP	SAVBUF0	;SEND ERRORS
SAVBUF7	JSR	CRLF
	JMP	SAVBUF6	;SEND ONE MORE CRLF AND EXIT
	
SAVEON	LDA	#1
	STA	SAVEF
	LDA	BINARY	; IF BINARY AND SAVING SET FLAG
	BEQ	SAVEON2
	ORA	#$80
	STA	BINARY
SAVEON2	RTS

SAVEOFF	LDA	#0
	STA	LOADF
	STA	SAVEF
	LDA	BINARY
	AND	#$7F
	STA	BINARY
	RTS


; INIT POINTERS, CLEAR BUFFER

INIT	JSR	RESPTR	; SET POINTERS
	LDY	#0
	LDA	#FILLCH
INIT1	STA	(NDX),Y	; SINCE WE ONLY SEND DATA READ CLEARING
	INY		; BUFFERS ISNT REALLY NEEDED BUT CAN
	BNE	INIT1	; HELP WITH TESTING
	INC	NDX+1
	DEC	CNTR
	BNE	INIT1
	JMP	RESPTR	; SET POINTERS

; RESET BUFFER POINTERS

RESPTR	LDA	#MAXPAG
	STA	CNTR
	LDA	#<BUFFER
	STA	NDX
	LDA	#>BUFFER
	STA	NDX+1
	RTS

; INIT DISK CONTROLLER PIA

INITPIA	LDY	#0
	STY	DISK+1	; SELECT DDRA
	LDA	#$40
	STA	DISK	; SET PORTA TO ALL INPUTS EXCEPT 6
	LDX	#4	; 0000 0100
	STX	DISK+1	; SELECT PORTA
	STA	DISK	; SET PA6 HIGH
	STY	DISK+3	; SELECT DDRB
	DEY
	STY	DISK+2	; SET PORTB TO ALL OUTPUTS
	STX	DISK+3	; SELECT PORTB
	STY	DISK+2	; SET PORTB OUTPUTS HIGH
	RTS

; STEP TO TRACK 0

TZERO	LDA	#$FB	; 1111 1011
	BNE	TZERO3	; DIR=INWARDS

TZERO1	LDA	#2	; 0000 0010
	BIT	DISK	; TEST 'TRK0' PIN
	BNE	TZERO2	; AT TRACK 0
	RTS

TZERO2	LDA	#$FF	; DIR=OUTWARDS (TO TRK0)

TZERO3	STA	DISK+2	; SET 'DIR' PIN
	JSR	DELAY2
	AND	#$F7	; 1111 0111
	STA	DISK+2	; SET 'STEP' PIN LOW
	JSR	DELAY2
	ORA	#8	; 0000 1000
	STA	DISK+2	; SET 'STEP' PIN HIGH
	LDX	#$30	; 60 MS
	JSR	DELAY
	BEQ	TZERO1	; ALWAYS

; STEP TO NEXT TRACK

TNEXT	LDA	#$FB	; 1111 1011
			; DIR=INWARDS
	STA	DISK+2	; SET 'DIR' PIN
	JSR	DELAY2
	AND	#$F7	; 1111 0111
	STA	DISK+2	; SET 'STEP' LOW
	JSR	DELAY2
	ORA	#8	; 0000 1000
	STA	DISK+2	; SET 'STEP' HIGH
	LDX	#$30	; 60MS
	JMP	DELAY

TPREV	LDA	#$FF	; 1111 1111
			; DIR=OUTWARDS
	STA	DISK+2	; SET 'DIR' PIN
	JSR	DELAY2
	AND	#$F7	; 1111 0111
	STA	DISK+2	; SET 'STEP' LOW
	JSR	DELAY2
	ORA	#8	; 0000 1000
	STA	DISK+2	; SET 'STEP' HIGH
	LDX	#$30	; 60MS
	JMP	DELAY

; READ CURRENT TRACK TO BUFFER

RDTRK	JSR	INIT
	LDX	#$7F	; 0111 1111
	STX	DISK+2	; LOAD DISK HEAD
	LDX	#$FF	; 320MS
	JSR	DELAY
RDTRK1	LDA	DISK	; WAIT START OF
	BMI	RDTRK1	; INDEX PULSE
RDTRK2	LDA	DISK	; WAIT END OF
	BPL	RDTRK2	; INDEX PULSE
	LDA	#3	; RESET ACIA
	STA	DISK+$10
	LDA	#$58	; DIVIDE BY 1, 8E1, RTS HIGH
	STA	DISK+$10
	LDY	#0	; READ UNTIL ALL PAGES DONE
	LDX	CNTR
; INPUT CHAR FROM DISK
RDTRK3	LDA	DISK	; TEST FOR INDEX PULSE
	BPL	RDTRK4
RDTRK3A	LDA	DISK+$10
	STA	(NDX),Y	; SAVE ERRORS
	LSR	A
	BCC	RDTRK3
	INY
	LDA	DISK+$11 ; READ BYTE
	STA	(NDX),Y
	INY
	BNE	RDTRK3
	INC	NDX+1
	DEX
	BNE	RDTRK3A	; SKIP INDEX PULSE TEST TO PREVENT
			; DATA OVERRUN
RDTRK4	LDA	#$FF
	STA	DISK+2	; UNLOAD DISK HEAD
	RTS

; DELAY 1.25MS PER LOOP

DELAY	LDY	#$F8
DELAY1	DEY
	BNE	DELAY1
	EOR	CNTR,X
	DEX
	BNE	DELAY
	RTS

DELAY2	JSR	DELAY21
DELAY21	RTS


; PRINT IN-LINE STRING

PRINT	PLA
	STA	NDX
	PLA
	STA	NDX+1
	LDY	#1
PRINT1	LDA	(NDX),Y
	BEQ	PRINT2
	JSR	OUTPUT
	INY
	BNE	PRINT1
PRINT2	TYA
	SEC
	ADC	NDX
	STA	NDX
	BCC	PRINT3
	INC	NDX+1
PRINT3	JMP	(NDX)

; PRINT HEX WORD (A,X)

PHEX	JSR	PHEXA
	TXA

; PRINT HEX BYTE (A)

PHEXA	BIT	BINARY
	BMI	PHEXA2
	PHA
	LSR	A
	LSR	A
	LSR	A
	LSR	A
	JSR	PHEXA1
	PLA
PHEXA1	AND	#$F
	ORA	#'0'
	CMP	#'9'+1
	BMI	PHEXA2
	CLC
	ADC	#7
PHEXA2	JMP	OUTPUT

; PRINT DECIMAL (A)

PDECA	TAX
	LDA	#0

; PRINT DECIMAL (A,X)

PDEC	STX	PDEC8
	STA	PDEC8+1
	LDA	#0
	LDX	#2
PDEC1	STA	PDEC10,X
	DEX
	BPL	PDEC1
	LDY	#16
	PHP
	SEI
	SED
PDEC2	ASL	PDEC8
	ROL	PDEC8+1
	LDX	#2
PDEC3	LDA	PDEC10,X
	ADC	PDEC10,X
	STA	PDEC10,X
	DEX
	BPL	PDEC3
	DEY
	BNE	PDEC2
	PLP
	LDX	#6
	LDA	#0
	STA	PDEC9
PDEC4	LDY	#3
	LDA	#0
PDEC5	ASL	PDEC10+2
	ROL	PDEC10+1
	ROL	PDEC10
	ROL	A
	DEY
	BPL	PDEC5
	TAY
	BNE	PDEC6
	CPX	#1
	BEQ	PDEC6
	LDY	PDEC9
	BEQ	PDEC7
PDEC6	INC	PDEC9
	ORA	#$30
	JSR	OUTPUT
PDEC7	DEX	
	BNE	PDEC4
	RTS

PDEC8	.WORD	0
PDEC9	.BYTE	0
PDEC10	.BYTE	0,0,0

CRLF	LDA	#CR
	JSR	OUTPUTC
	LDA	#LF
	JMP	OUTPUTC

ANYKEY	JSR	PRINT
	.BYTE	CR,LF
	.BYTE	'PRESS ANY KEY WHEN READY',CR,LF,0
	JMP	INPUT

READMSG	PHA
	JSR	PRINT
	.BYTE	CR,LF
	.BYTE	'READING TRACK ',0
	PLA
	JMP	PDECA

SAVMSG	JSR	PRINT
	.BYTE	' ...SAVING',0
	RTS

RESERR	LDA	#0
	STA	ERRCNT
	STA	ERRCNT+1
	RTS

SOUTC   PHA
SOUTCL  LDA 	$FC00
        LSR 	A
        LSR 	A
        BCC 	SOUTCL
        PLA
        STA 	$FC01
	LDX	OUTXHLD
        RTS

OUTPUTC	BIT	BINARY	; MUST NOT MODIFY Y
	BMI	OUTRTS
OUTPUT	STX	OUTXHLD
	LDX	SAVEF
	BNE	SOUTC
	CMP	#CR
	BEQ	DOCR
	CMP	#LF
	BEQ	DOLF
	LDX	SCRLOC
	STA	$D740,X
	INX
	STX 	SCRLOC
	LDX	OUTXHLD
OUTRTS	RTS
OUTXHLD	.BYTE 0
DOCR	LDA	#0
	STA	SCRLOC
	LDX	OUTXHLD
	RTS
DOLF	TYA
	PHA
	LDA	#$20	;Clear last line we aren't using
	LDX	#$80	;So content in it won't scroll up
CLRLN	STA	$D700,X
	INX
	BNE	CLRLN
	LDY	#$CF
SETNXT	INY		; OS65D Scroll routine
	STY	MOVE+2
	STY	MOVE+5
MOVE	LDA	$D040,X
	STA	$D000,X
	INX
	BEQ	SETNXT
	BPL	MOVE
	CPY	#$D7
	BCC	MOVE
	PLA
	TAY
	LDX	OUTXHLD
	RTS
	
GETTRK	JSR	PRINT
	.BYTE	CR,LF
	.BYTE	'ENTER TRACK ## TO START DUMP AT ',0
	JSR	INPUT
	STA	TRK
	JSR	OUTPUT
	LDA	TRK
	SEC
	SBC	#'0'
	ASL	A
	STA	CNTR
	ASL	A
	ASL	A
	CLC
	ADC	CNTR	; MULTIPLY BY 10
	STA	TRK
	JSR	INPUT
	PHA
	JSR	OUTPUT
	JSR	CRLF
	PLA
	SEC
	SBC	#'0'
	CLC
	ADC	TRK
	STA	TRK
	CMP	MAXTRK
	BMI	GETTEND
TRKERR	JSR	PRINT
	.BYTE	'TRACK NUMBER INVALID ',CR,LF,0
	JMP	GETTRK
GETTEND	JSR	PRINT
	.BYTE	CR,LF
	.BYTE	'ENTER TRACK ## TO END DUMP AT ',0
	JSR	INPUT
	STA	ENDTRK
	JSR	OUTPUT
	LDA	ENDTRK
	SEC
	SBC	#'0'
	ASL	A
	STA	CNTR
	ASL	A
	ASL	A
	CLC
	ADC	CNTR	; MULTIPLY BY 10
	STA	ENDTRK
	JSR	INPUT
	PHA
	JSR	OUTPUT
	JSR	CRLF
	PLA
	SEC
	SBC	#'0'
	CLC
	ADC	ENDTRK
	STA	ENDTRK
	CMP	MAXTRK
	BMI	GETRET
	JMP	TRKERR
GETRET	RTS

SENT	.BYTE	0	; SENT DATA FLAG
ERRDMP	.BYTE	0	; 0 = SEND DATA, FF = SEND ERRORS
CNT	.WORD	0	; DATA READ COUNTER
ERRCNT	.WORD	0	; ERROR COUNTER
TRK	.BYTE	0	; CURRENT TRACK
ENDTRK	.BYTE	0	; LAST TRACK
SCRLOC	.BYTE	0	: SCREEN CHARACTER COUNTER
BINARY	.BYTE	1	; 1 = BINARY, $81 = BINARY WITH SAVEON SET
DRVNUM	.BYTE	0	; DRIVE NUMBER, 0 = A
MAXTRK	.BYTE	77	; # TRACKS TO READ 77 8", 40 5 1/4"
READCNT	.BYTE	0	; NON ZERO TO REREAD TRACK

	.END
